# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`prometheus_self.py`
Prometheus Self exporter class.
"""
import logging

from flask import current_app, make_response
from prometheus_client import Gauge, generate_latest

from .base import MetricsExporter

LOGGER = logging.getLogger(__name__)


class PrometheusSelfExporter(MetricsExporter):
    """Prometheus exporter class which uses
    the API mode of itself to export the metrics."""

    REGISTERED_URL: bool = False
    REGISTERED_METRICS: dict = {}

    def __init__(self):
        if not self.REGISTERED_URL:
            current_app.add_url_rule("/metrics", view_func=self.serve_metrics)
            PrometheusSelfExporter.REGISTERED_URL = True

    @staticmethod
    def serve_metrics():
        """Serves prometheus metrics

        Returns:
            object: Flask HTTP Response
        """
        resp = make_response(generate_latest(), 200)
        resp.mimetype = "text/plain"
        return resp

    def export_metric(self, data):
        """Export data to Prometheus global registry.

        Args:
            data (dict): Metric data.
        """
        name = data["name"]
        description = data["description"]
        value = data["value"]

        # Write timeseries w/ metric labels.
        labels = data["labels"]
        gauge = self.REGISTERED_METRICS.get(name)
        if gauge is None:
            gauge = Gauge(
                name,
                description,
                labelnames=labels.keys(),
            )
            PrometheusSelfExporter.REGISTERED_METRICS[name] = gauge
        gauge.labels(*labels.values()).set(value)



# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`utils.py`
Utility functions.
"""
import argparse
import errno
import importlib
import logging
import os
import pprint
import re
import sys
import warnings
from collections.abc import Mapping
from datetime import datetime
from pathlib import Path
from typing import Optional

import yaml
from dateutil import tz

from slo_generator.constants import DEBUG

try:
    # pytype: disable=import-error
    from google.cloud import storage  # type: ignore[attr-defined]

    GCS_ENABLED = True
except ImportError:
    GCS_ENABLED = False

LOGGER = logging.getLogger(__name__)


# pylint: disable=dangerous-default-value
def load_configs(
    path: str, ctx: os._Environ = os.environ, kind: Optional[str] = None
) -> list:
    """Load multiple slo-generator configs from a folder path.

    Args:
        path (str): Folder path.
        ctx (dict): Context for variable replacement.
        kind (str): Config kind filter.

    Returns:
        list: List of configs downloaded and parsed.
    """
    configs = [
        load_config(str(p), ctx=ctx, kind=kind)
        for p in sorted(Path(path).glob("*.yaml"))
    ]
    return [cfg for cfg in configs if cfg]


# pylint: disable=dangerous-default-value
def load_config(
    path: str, ctx: os._Environ = os.environ, kind: Optional[str] = None
) -> Optional[dict]:
    """Load any slo-generator config, from a local path, a GCS URL, or directly
    from a string content.

    Args:
        path (str): GCS URL, file path, or data as string.
        ctx (dict): Context for variable replacement.
        kind (str): Config kind filter.

    Returns:
        dict: Config downloaded and parsed.
    """
    abspath = Path(path)
    try:
        if path.startswith("gs://"):
            if not GCS_ENABLED:
                warnings.warn(
                    "To load a file from GCS, you need `google-cloud-storage` "
                    "installed. Please install it using pip by running "
                    "`pip install google-cloud-storage`",
                    ImportWarning,
                )
                sys.exit(1)
            config = parse_config(content=download_gcs_file(str(path)), ctx=ctx)
        elif abspath.is_file():
            config = parse_config(path=str(abspath.resolve()), ctx=ctx)
        else:
            LOGGER.debug(f"Path {path} not found. Trying to load from string")
            config = parse_config(content=str(path), ctx=ctx)

        # Filter on 'kind'
        if kind and (not isinstance(config, dict) or kind != config.get("kind", "")):
            config = None
        return config

    except OSError as exc:
        if exc.errno == errno.ENAMETOOLONG:
            return parse_config(content=str(path), ctx=ctx)
        raise


# pylint: disable=dangerous-default-value
def parse_config(
    path: Optional[str] = None, content=None, ctx: os._Environ = os.environ
):
    """Load a yaml configuration file and resolve environment variables in it.

    Args:
        path (str): the path to the yaml file.
        content (str): the config content as a dict string.
        ctx (dict): Context to replace env variables from (defaults to
            `os.environ`).

    Returns:
        dict: Parsed YAML dictionary.
    """
    pattern = re.compile(r".*?\${(\w+)}.*?")

    def replace_env_vars(content, ctx) -> str:
        """Replace env variables in content from context.

        Args:
            content (str): String to parse.
            ctx (dict): Context to replace vars from.

        Returns:
            str: the parsed string with the env var replaced.
        """
        match = pattern.findall(content)
        if match:
            full_value = content
            for var in match:
                try:
                    full_value = full_value.replace(f"${{{var}}}", ctx[var])
                except KeyError as exception:
                    LOGGER.error(
                        f'Environment variable "{var}" should be set.', exc_info=True
                    )
                    raise exception
            content = full_value
        return content

    if path:
        with Path(path).open(encoding="utf8") as config:
            content = config.read()
    if ctx:
        content = replace_env_vars(content, ctx)
    data = yaml.safe_load(content)
    if isinstance(data, str):
        error = (
            "Error serializing config into dict. This might be due to a syntax "
            "error in the YAML / JSON config file."
        )
        LOGGER.error(error)

    LOGGER.debug(pprint.pformat(data))
    return data


def setup_logging():
    """Setup logging for the CLI."""
    if DEBUG == 1:
        print(f"DEBUG mode is enabled. DEBUG={DEBUG}")
        level = logging.DEBUG
        format_str = "%(name)s - %(levelname)s - %(message)s"
    else:
        level = logging.INFO
        format_str = "%(levelname)s - %(message)s"
    logging.basicConfig(
        stream=sys.stdout, level=level, format=format_str, datefmt="%m/%d/%Y %I:%M:%S"
    )
    logging.getLogger("googleapiclient").setLevel(logging.ERROR)

    # Ignore Cloud SDK warning when using a user instead of service account
    try:
        # pylint: disable=import-outside-toplevel
        from google.auth._default import _CLOUD_SDK_CREDENTIALS_WARNING

        warnings.filterwarnings("ignore", message=_CLOUD_SDK_CREDENTIALS_WARNING)
    except ImportError:
        pass


def get_human_time(timestamp: int, timezone: Optional[str] = None) -> str:
    """Get human-readable timestamp from UNIX UTC timestamp.

    Args:
        timestamp (int): UNIX UTC timestamp.
        timezone (optional): Explicit timezone (e.g: "America/Chicago").

    Returns:
        str: Formatted human-readable date in ISO format.

    Example:
        >>> get_human_time(1702660513.987654, timezone='Europe/Paris')
        >>> 2023-12-15T18:15:13.987654+01:00
        which corresponds to the timestamp in ISO format
    """
    if timezone is not None:  # get timezone from arg
        to_zone = tz.gettz(timezone)
    else:  # auto-detect locale
        to_zone = tz.tzlocal()

    # NB: as we set tz, dt_tz is an "aware" datetime
    # see: https://docs.python.org/3/library/datetime.html#aware-and-naive-objects
    dt_tz = datetime.fromtimestamp(timestamp, tz=to_zone)
    date_str = dt_tz.isoformat()

    return date_str


def get_exporters(config: dict, spec: dict) -> list:
    """Get SLO exporters configs from spec and global config.

    Args:
        config (dict): Global config.
        spec (dict): SLO config.

    Returns:
        list: List of dict containing exporters configurations.
    """
    all_exporters = config.get("exporters", {})
    spec_exporters = spec.get("exporters", [])
    exporters = []
    for exporter in spec_exporters:
        if exporter not in all_exporters.keys():
            LOGGER.error(f'Exporter "{exporter}" not found in config.')
            continue
        exporter_data = all_exporters[exporter]
        exporter_data["name"] = exporter
        if "." in exporter:  # support custom exporter
            exporter_data["class"] = exporter
        else:  # core exporter
            exporter_data["class"] = capitalize(snake_to_caml(exporter.split("/")[0]))
        exporters.append(exporter_data)
    return exporters


def get_backend(config: dict, spec: dict):
    """Get SLO backend config from spec and global config.

    Args:
        config (dict): Global config.
        spec (dict): SLO config.

    Returns:
        list: List of dict containing exporters configurations.
    """
    all_backends = config.get("backends", {})
    spec_backend = spec["backend"]
    backend_data = {}
    if spec_backend not in all_backends.keys():
        LOGGER.error(f'Backend "{spec_backend}" not found in config. Exiting.')
        sys.exit(0)
    backend_data = all_backends[spec_backend]
    backend_data["name"] = spec_backend
    if "." in spec_backend:  # custom backend
        backend_data["class"] = spec_backend
    else:  # built-in backend
        backend_data["class"] = capitalize(snake_to_caml(spec_backend.split("/")[0]))
    return backend_data


def get_error_budget_policy(config: dict, spec: dict):
    """Get error budget policy from spec and global config.

    Args:
        config (dict): Global config.
        spec (dict): SLO config.

    Returns:
        list: List of dict containing exporters configurations.
    """
    all_ebp = config.get("error_budget_policies", {})
    spec_ebp = spec.get("error_budget_policy", "default")
    if spec_ebp not in all_ebp.keys():
        LOGGER.error(f'Error budget policy "{spec_ebp}" not found in config. Exiting.')
        sys.exit(0)
    return all_ebp[spec_ebp]


def get_backend_cls(backend: str):
    """Get backend class.

    Args:
        backend (str): Exporter type.

    Returns:
        class: Backend class.
    """
    expected_type = "Backend"
    return import_cls(backend, expected_type)


def get_exporter_cls(exporter: str):
    """Get exporter class.

    Args:
        exporter (str): Backend type.

    Returns:
        class: Exporter class.
    """
    expected_type = "Exporter"
    return import_cls(exporter, expected_type)


def import_cls(cls_name, expected_type):
    """Import class or method dynamically from full name.
    If `cls_name` is not part of the core, try import from local path (plugins).

    Args:
        cls_name: Class name to import.
        expected_type: Type of class expected.

    Returns:
        obj: Imported class or method object.
    """
    # plugin class
    if "." in cls_name:
        package, name = cls_name.rsplit(".", maxsplit=1)
        return import_dynamic(package, name, prefix=expected_type)

    # slo-generator core class
    modules_name = f"{expected_type.lower()}s"
    full_cls_name = f"{cls_name}{expected_type}"
    filename = re.sub(r"(?<!^)(?=[A-Z])", "_", cls_name).lower()
    return import_dynamic(
        f"slo_generator.{modules_name}.{filename}", full_cls_name, prefix=expected_type
    )


def import_dynamic(package: str, name: str, prefix: str = "class"):
    """Import class or method dynamically from package and name.

    Args:
        package: Where the method or class is located in the import path.
        name: Name of method or class.

    Returns:
        obj: Imported class or method object.
    """
    try:
        return getattr(importlib.import_module(package), name)
    except Exception as exception:  # pylint: disable=W0703
        dep = package.split(".")[-1]
        warnings.warn(
            f'{prefix} "{package}.{name}" not found.\nPlease ensure that:\n'
            f"1. Package and class name are valid.\n"
            f"2. Extra dependency {dep} is installed. If not, install it "
            f'locally with "pip install slo-generator[{dep}]" or remotely '
            f'by adding "slo-generator[{dep}]" to your requirements.txt.',
            ImportWarning,
        )
        if DEBUG:
            LOGGER.debug(exception, exc_info=True)
        return None


def capitalize(word: str) -> str:
    """Only capitalize the first letter of a word, even when written in
    CamlCase.

    Args:
        word (str): Input string.

    Returns:
        str: Input string with first letter capitalized.
    """
    return re.sub("([a-zA-Z])", lambda x: x.groups()[0].upper(), word, 1)


def snake_to_caml(word: str) -> str:
    """Convert a string written in snake_case to a string in CamlCase.

    Args:
        word (str): Input string.

    Returns:
        str: Output string.
    """
    return re.sub("_.", lambda x: x.group()[1].upper(), word)


def caml_to_snake(word: str) -> str:
    """Convert a string written in CamlCase to a string written in snake_case.

    Args:
        word (str): Input string.

    Returns:
        str: Output string.
    """
    return re.sub(r"(?<!^)(?=[A-Z])", "_", word).lower()


def dict_snake_to_caml(data: dict) -> dict:
    """Convert dictionary with keys written in snake_case to another one with
    keys written in CamlCase.

    Args:
        data (dict): Input dictionary.

    Returns:
        dict: Output dictionary.
    """
    return apply_func_dict(data, snake_to_caml)


def apply_func_dict(data: dict, func) -> dict:
    """Apply function on a dictionary keys.

    Args:
        data (dict): Input dictionary.

    Returns:
        dict: Output dictionary.
    """
    if isinstance(data, Mapping):
        return {func(k): apply_func_dict(v, func) for k, v in data.items()}
    return data


def str2bool(string: str) -> bool:
    """Convert a string to a boolean.

    Args:
        string (str): String to convert

    Returns:
        bool: Boolean value.

    Raises:
        `argparse.ArgumentTypeError`: IF no acceptable boolean string is found.
    """
    if isinstance(string, bool):
        return string
    if string.lower() in ("yes", "true", "t", "y", "1"):
        return True
    if string.lower() in ("no", "false", "f", "n", "0"):
        return False
    raise argparse.ArgumentTypeError("Boolean value expected.")


def download_gcs_file(url: str) -> dict:
    """Download config from GCS.

    Args:
        url: Config URL.

    Returns:
        dict: Loaded configuration.
    """
    client = storage.Client()
    bucket, filepath = decode_gcs_url(url)
    bucket = client.get_bucket(bucket)
    blob = bucket.blob(filepath)
    return blob.download_as_string(client=None).decode("utf-8")


def decode_gcs_url(url: str) -> tuple:
    """Decode GCS URL.

    Args:
        url (str): GCS URL.

    Returns:
        tuple: (bucket_name, file_path)
    """
    split_url = url.split("/")
    bucket_name = split_url[2]
    file_path = "/".join(split_url[3:])
    return (bucket_name, file_path)


# pylint: disable=dangerous-default-value
def get_files(source, extensions=["yaml", "yml", "json"]) -> list:
    """Get all files matching extensions.

    Args:
        extensions (list): List of extensions to match.

    Returns:
        list: List of all files matching extensions relative to source folder.
    """
    all_files: list = []
    for ext in extensions:
        all_files.extend(Path(source).rglob(f"*.{ext}"))
    return all_files


def get_target_path(source_dir, target_dir, relative_path, mkdir: bool = True):
    """Get target file path from a source directory, a target directory and a
    path relative to the source directory.

    Args:
        source_dir (Path): path to source directory.
        target_dir (pathlib.Path): path to target directory.
        relative_path (pathlib.Path): path relative to source directory.
        mkdir (bool): Create directory structure for target path.

    Returns:
        pathlib.Path: path to target file.
    """
    source_dir = source_dir.resolve()
    target_dir = target_dir.resolve()
    relative_path = relative_path.relative_to(source_dir)
    common_path = os.path.commonpath([source_dir, target_dir])
    target_path = common_path / target_dir.relative_to(common_path) / relative_path
    if mkdir:
        target_path.parent.mkdir(parents=True, exist_ok=True)
    return target_path


def fmt_traceback(exc) -> str:
    """Format exception to be human-friendly.

    Args:
        exc (Exception): Exception to format.

    Returns:
        str: Formatted exception.
    """
    return exc.__class__.__name__ + ": " + str(exc).replace("\n", " ")



# Copyright 2021 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`v1tov2.py`
Migrate utilities for migrating slo-generator configs from v1 to v2.
"""
# pylint: disable=line-too-long, too-many-statements, too-many-ancestors, too-many-locals, too-many-nested-blocks, unused-argument
# flake8: noqa
# pytype: skip-file
import copy
import itertools
import pprint
import random
import string
import sys
from collections import OrderedDict
from pathlib import Path
from typing import Optional

import click
from ruamel import yaml

from slo_generator import utils
from slo_generator.constants import (
    BOLD,
    CONFIG_SCHEMA,
    ENDC,
    FAIL,
    GREEN,
    METRIC_LABELS_COMPAT,
    METRIC_METADATA_LABELS_TOP_COMPAT,
    PROVIDERS_COMPAT,
    RED,
    RIGHT_ARROW,
    SLO_CONFIG_SCHEMA,
    SUCCESS,
    WARNING,
)

yaml.explicit_start = True  # type: ignore[attr-defined]
yaml.default_flow_style = None  # type: ignore[attr-defined]
yaml.preserve_quotes = True  # type: ignore[attr-defined]


# pylint: disable=too-many-arguments
def do_migrate(
    source,
    target,
    error_budget_policy_path: list,
    exporters_path: list,
    version: str,
    quiet: bool = False,
    verbose: int = 0,
):
    """Process all SLO configs in folder and generate new SLO configurations.

    Args:
        source (str): Source SLO configs folder.
        target (str): Target SLO configs folder.
        error_budget_policy_path (list): Error budget policy paths.
        exporters_path (list): Exporters paths.
        glob (str): Glob expression to add to source path.
        version (str): slo-generator major version string (e.g: v1, v2, ...)
        quiet (bool, optional): If true, do not prompt for user input.
        verbose (int, optional): Verbose level.
    """
    curver: str = "v1"
    shared_config = CONFIG_SCHEMA
    cwd = Path.cwd()
    source = Path(source).resolve()
    target = Path(target).resolve()
    source_str = source.relative_to(cwd)  # human-readable path
    target_str = target.relative_to(cwd)  # human-readable path
    ebp_paths = [Path(ebp) for ebp in error_budget_policy_path]
    exporters_paths = [Path(exp) for exp in exporters_path]

    # Create target folder if it doesn't exist
    target.mkdir(parents=True, exist_ok=True)

    # Translate error budget policy to v2 and put into shared config
    if ebp_paths:
        ebp_func = getattr(sys.modules[__name__], f"ebp_{curver}to{version}")
        ebp_func(
            ebp_paths,
            shared_config=shared_config,
            quiet=quiet,
        )

    # Translate exporters to v2 and put into shared config
    if exporters_paths:
        exporters_func = getattr(
            sys.modules[__name__], f"exporters_{curver}to{version}"
        )
        exp_keys = exporters_func(
            exporters_paths,
            shared_config=shared_config,
            quiet=quiet,
        )

    # Process SLO configs
    click.secho("=" * 50)
    click.secho(
        f"Migrating slo-generator configs to {version} ...", fg="cyan", bold=True
    )
    paths = utils.get_files(source)
    if not paths:
        click.secho(f"{FAIL} No SLO configs found in {source}", fg="red", bold=True)
        sys.exit(1)

    curver = ""
    for source_path in paths:
        if source_path in ebp_paths + exporters_paths:
            continue
        source_path_str = source_path.relative_to(cwd)
        target_path = utils.get_target_path(
            source,
            target,
            source_path,
            mkdir=True,
        )
        target_path_str = target_path.resolve().relative_to(cwd)
        slo_config_str = source_path.open().read()
        slo_config, ind, blc = yaml.util.load_yaml_guess_indent(slo_config_str)
        curver = detect_config_version(slo_config)
        if not curver:
            continue

        # Source path info
        click.secho("-" * 50)
        click.secho(f"{WARNING}{source_path_str}{ENDC} [{curver}] ")

        # If config version is same as target version, continue
        if curver == version:
            click.secho(
                f"{FAIL} {source_path_str} is already in {version} format",
                fg="red",
                bold=True,
            )
            continue

        # Create target dirs if needed
        target_path.parent.mkdir(parents=True, exist_ok=True)

        # Run vx to vy migrator method
        slo_func = getattr(sys.modules[__name__], f"slo_config_{curver}to{version}")
        slo_config_v2 = slo_func(
            slo_config,
            shared_config=shared_config,
            shared_exporters=exp_keys if exporters_paths else [],
            quiet=quiet,
        )
        if not slo_config_v2:
            continue

        # Write resulting config to target path
        extra = "(replaced)" if target_path_str == source_path_str else ""
        click.secho(f"{RIGHT_ARROW} {GREEN}{target_path_str}{ENDC} [{version}] {extra}")
        with target_path.open("w", encoding="utf8") as conf:
            yaml.round_trip_dump(
                slo_config_v2,
                conf,
                indent=ind,
                block_seq_indent=blc,
                default_flow_style=None,
            )
        click.secho(f"{SUCCESS} Success !", fg="green", bold=True)

    # Write shared config to file
    click.secho("=" * 50)
    shared_config_path = target / "config.yaml"
    shared_config_path_str = shared_config_path.relative_to(cwd)
    with shared_config_path.open("w", encoding="utf8") as conf:
        click.secho(
            f"Writing slo-generator config to {shared_config_path_str} ...",
            fg="cyan",
            bold=True,
        )
        yaml.round_trip_dump(
            shared_config,
            conf,
            Dumper=CustomDumper,
            indent=2,
            block_seq_indent=0,
            explicit_start=True,
        )
        click.secho(f"{SUCCESS} Success !", fg="green", bold=True)

    # Remove error budget policy file
    # click.secho('=' * 50)
    # click.secho(f'Removing {error_budget_policy_path} ...',
    #             fg='cyan',
    #             bold=True)
    # error_budget_policy_path.unlink()
    # click.secho(f'{SUCCESS} Success !', fg='green', bold=True)

    # Print next steps
    relative_ebp_path = ebp_paths[0].relative_to(cwd)
    click.secho("=" * 50)
    click.secho(
        f"\n{SUCCESS} Migration of `slo-generator` configs to v2 completed successfully ! Configs path: {target_str}/.\n",
        fg="green",
        bold=True,
    )
    click.secho("=" * 50)
    click.secho(
        f"{BOLD}PLEASE FOLLOW THE MANUAL STEPS BELOW TO FINISH YOUR MIGRATION:",
        fg="red",
        bold=True,
    )
    click.secho(
        f"""
    1 - Commit the updated SLO configs and your shared SLO config to version control.
    2 - [local/k8s/cloudbuild] Update your slo-generator command:
    {RED}  [-] slo-generator -f {source_str} -b {relative_ebp_path}{ENDC}
    {GREEN}  [+] slo-generator -f {target_str} -c {target_str}/config.yaml{ENDC}
    """
    )
    # 3 - [terraform] Upgrade your `terraform-google-slo` modules:
    # 3.1 - Upgrade the module `version` to 2.0.0.
    # 3.2 - Replace `error_budget_policy` field in your `slo` and `slo-pipeline` modules by `shared_config`
    # 3.3 - Replace `error_budget_policy.yaml` local variable to `config.yaml`


# pylint: disable=dangerous-default-value
def exporters_v1tov2(
    exporters_paths: list, shared_config: dict = {}, quiet: bool = False
) -> list:
    """Translate exporters to v2 and put into shared config.

    Args:
        exporters_path (list): List of exporters file paths.
        shared_config (dict): Shared config to add exporters to.
        quiet (bool): Quiet mode.

    Returns:
        list: List of exporters keys added to shared config.
    """
    exp_keys = []
    for exp_path in exporters_paths:
        with open(exp_path, encoding="utf-8") as conf:
            # pylint: disable=E1111
            content = yaml.load(conf, Loader=yaml.SafeLoader)
        exporters = content

        # If exporters file has sections, concatenate all of them
        if isinstance(content, dict):
            exporters = []
            for _, value in content.items():
                exporters.extend(value)

        # If exporter not in general config, add it and add an alias for the
        # exporter. Refer to the alias in the SLO config file.
        for exporter in exporters:
            exporter = OrderedDict(exporter)
            exp_key = add_to_shared_config(
                exporter, shared_config, "exporters", quiet=quiet
            )
            exp_keys.append(exp_key)
    return exp_keys


# pylint: disable=dangerous-default-value
def ebp_v1tov2(ebp_paths: list, shared_config: dict = {}, quiet: bool = False) -> list:
    """Translate error budget policies to v2 and put into shared config

    Args:
        ebp_paths (list): List of error budget policies file paths.
        shared_config (dict): Shared config to add exporters to.
        quiet (bool): Quiet mode.

    Returns:
        list: List of error budget policies keys added to shared config.
    """
    ebp_keys = []
    for ebp_path in ebp_paths:
        with open(ebp_path, encoding="utf-8") as conf:
            # pylint: disable=E1111
            error_budget_policy = yaml.load(conf, Loader=yaml.SafeLoader)
        # pylint: disable=E1133
        for step in error_budget_policy:
            step["name"] = step.pop("error_budget_policy_step_name")
            step["burn_rate_threshold"] = step.pop("alerting_burn_rate_threshold")
            step["alert"] = step.pop("urgent_notification")
            step["message_alert"] = step.pop("overburned_consequence_message")
            step["message_ok"] = step.pop("achieved_consequence_message")
            step["window"] = step.pop("measurement_window_seconds")

        ebp = {"steps": error_budget_policy}
        if ebp_path.name == "error_budget_policy.yaml":
            ebp_key = "default"
        else:
            ebp_key = ebp_path.stem.replace("error_budget_policy_", "")
        ebp_key = add_to_shared_config(
            ebp,
            shared_config,
            "error_budget_policies",
            ebp_key,
            quiet=quiet,
        )
        ebp_keys.append(ebp_key)
    return ebp_keys


# pylint: disable=dangerous-default-value
def slo_config_v1tov2(
    slo_config: dict,
    shared_config: dict = {},
    shared_exporters: list = [],
    quiet: bool = False,
    verbose: int = 0,
):
    """Process old SLO config v1 and generate SLO config v2.

    Args:
        slo_config (dict): SLO Config v1.
        shared_config (dict): SLO Generator config.
        shared_exporters (list): Shared exporters keys to add to SLO configs.
        quiet (bool): If true, do not ask for user input.
        verbose (int): Verbose level.

    Returns:
        dict: SLO Config v2.
    """
    # SLO config v2 skeleton
    slo_config_v2 = OrderedDict(copy.deepcopy(SLO_CONFIG_SCHEMA))
    slo_config_v2["apiVersion"] = "sre.google.com/v2"
    slo_config_v2["kind"] = "ServiceLevelObjective"
    missing_keys = [
        key
        for key in ["service_name", "feature_name", "slo_name", "backend"]
        if key not in slo_config
    ]
    if missing_keys:
        click.secho(
            f"Invalid SLO configuration: missing key(s) {missing_keys}.", fg="red"
        )
        return None

    # Get fields from old config
    slo_metadata_name_fmt = "{service_name}-{feature_name}-{slo_name}"
    slo_metadata_name = slo_metadata_name_fmt.format(**slo_config)
    slo_description = slo_config.pop("slo_description")
    slo_target = slo_config.pop("slo_target")
    service_level_indicator = slo_config["backend"].pop("measurement", {})
    backend = slo_config["backend"]
    method = backend.pop("method")
    exporters = slo_config.get("exporters", [])
    if isinstance(exporters, dict):  # single exporter, deprecated
        exporters = [exporters]

    # Fill spec
    slo_config_v2["metadata"]["name"] = slo_metadata_name
    slo_config_v2["metadata"]["labels"] = {
        "service_name": slo_config["service_name"],
        "feature_name": slo_config["feature_name"],
        "slo_name": slo_config["slo_name"],
    }
    other_labels = {
        k: v
        for k, v in slo_config.items()
        if k not in ["service_name", "feature_name", "slo_name", "backend", "exporters"]
    }
    slo_config_v2["metadata"]["labels"].update(other_labels)
    slo_config_v2["spec"]["description"] = slo_description
    slo_config_v2["spec"]["goal"] = slo_target

    # Process backend
    backend = OrderedDict(backend)
    backend_key = add_to_shared_config(
        backend,
        shared_config,
        "backends",
        quiet=quiet,
    )
    slo_config_v2["spec"]["backend"] = backend_key
    slo_config_v2["spec"]["method"] = method

    # If exporter not in general config, add it and add an alias for the
    # exporter. Refer to the alias in the SLO config file.
    for exporter in exporters:
        exporter = OrderedDict(exporter)
        exp_key = add_to_shared_config(
            exporter, shared_config, "exporters", quiet=quiet
        )
        slo_config_v2["spec"]["exporters"].append(exp_key)

    # Add shared exporters to slo config
    for exp_key in shared_exporters:
        slo_config_v2["spec"]["exporters"].append(exp_key)

    # Fill spec
    slo_config_v2["spec"]["service_level_indicator"] = service_level_indicator

    if verbose > 0:
        pprint.pprint(dict(slo_config_v2))
    return dict(slo_config_v2)


def report_v2tov1(report: dict) -> dict:
    """Convert SLO report from v2 to v1 format, for exporters to be
    backward-compatible with v1 data format.

    Args:
        report (dict): SLO report.

    Returns:
        dict: Converted SLO report.
    """
    mapped_report: dict = {}
    for key, value in report.items():
        # If a metadata label is passed, use the metadata label mapping
        if key == "metadata":
            mapped_report["metadata"] = {}
            for subkey, subvalue in value.items():
                # v2 `metadata.labels` attributes map to `metadata` attributes
                # in v1
                if subkey == "labels":
                    labels = subvalue
                    for labelkey, labelval in labels.items():
                        # Top-level labels like 'service_name', 'feature_name',
                        # and 'slo_name'.
                        if labelkey in METRIC_METADATA_LABELS_TOP_COMPAT:
                            mapped_report[labelkey] = labelval

                        # Other labels that are mapped to 'metadata' in the v1
                        # report
                        else:
                            mapped_report["metadata"][labelkey] = labelval

                # ignore the name attribute which is just a concatenation of
                # service_name, feature_name and slo_name
                elif subkey == "name":
                    continue

                # other metadata labels are still mapped to the v1 `metadata`
                # attributes
                else:
                    mapped_report["metadata"][subkey] = subvalue

        # If a key in the default label mapping is passed, use the default
        # label mapping
        elif key in METRIC_LABELS_COMPAT:
            mapped_report[METRIC_LABELS_COMPAT[key]] = value
        else:
            mapped_report[key] = value
    return mapped_report


def get_random_suffix() -> str:
    """Get random suffix for our backends / exporters when configs clash."""
    return "".join(random.choices(string.digits, k=4))  # nosec B311


def add_to_shared_config(
    new_obj: dict, shared_config: dict, section: str, key=None, quiet: bool = False
):
    """Add an object to the shared_config.

    If the object with the same config already exists in the shared config,
    simply return its key.

    If the object does not exist in the shared config:
    * If the default key is already taken, add a random suffix to it.
    * If the default key is not taken, add the new object to the config.

    Args:
        new_obj (OrderedDict): Object to add to shared_config.
        shared_config (dict): Shared config to add object to.
        section (str): Section name in shared config to add the object under.
        key (str): Key if cannot be infered.
        quiet (bool): If True, do not ask for user input.

    Returns:
        str: Object key in the shared config.
    """
    shared_obj = shared_config[section]
    key = key or new_obj.pop("class", None)
    if not key:
        raise ValueError("Object key is undefined.")
    if "." not in key:
        key = utils.caml_to_snake(PROVIDERS_COMPAT.get(key, key))

    existing_obj = {
        k: v
        for k, v in shared_obj.items()
        if k.startswith(key.split("/")[0]) and str(v) == str(dict(new_obj))
    }
    if existing_obj:
        key = next(iter(existing_obj))
        # click.secho(f'Found existing {section} {key}')
    else:
        if key in shared_obj.keys():  # key conflicts
            if quiet:
                key += "/" + get_random_suffix()
            else:
                name = section.rstrip("s")
                cfg = pprint.pformat({key: dict(new_obj)})
                valid = False
                while not valid:
                    click.secho(
                        f"\nNew {name} found with the following config:\n{cfg}",
                        fg="cyan",
                        blink=True,
                    )
                    user_input = click.prompt(
                        f"\n{RED}{BOLD}Please give this {name} a name{ENDC}", type=str
                    )
                    former_key = key
                    key += "/" + user_input.lower()
                    if key in shared_obj.keys():
                        click.secho(
                            f'{name.capitalize()} "{key}" already exists in shared config',
                            fg="red",
                            bold=True,
                        )
                        key = former_key
                    else:
                        valid = True
                click.secho(
                    f"Backend {key} was added to shared config.", fg="green", bold=True
                )

        # click.secho(f"Adding new {section} {key}")
        shared_obj[key] = dict(new_obj)
        shared_config[section] = dict(sorted(shared_obj.items()))
    return key


def detect_config_version(config: dict) -> str:
    """Return version of an slo-generator config based on the format.

    Args:
        config (dict): slo-generator configuration.

    Returns:
        str: SLO config version.
    """
    if not isinstance(config, dict):
        click.secho(
            "Config does not correspond to any known SLO config versions.", fg="red"
        )
        return None
    api_version: str = config.get("apiVersion", "")
    kind = config.get("kind", "")
    if not kind:  # old v1 format
        return "v1"
    return api_version.split("/")[-1]


def peek(iterable):
    """Check if iterable is empty.

    Args:
        iterable (collections.Iterable): an iterable

    Returns:
        iterable (collections.Iterable): the same iterable, or None if empty.
    """
    try:
        first = next(iterable)
    except StopIteration:
        return None
    return first, itertools.chain([first], iterable)


# pylint: disable=too-few-public-methods
class CustomDumper(yaml.RoundTripDumper):
    """Dedicated YAML dumper to insert lines between top-level objects.

    Args:
        data (str): Line data.
    """

    # HACK: insert blank lines between top-level objects
    # inspired by https://stackoverflow.com/a/44284819/3786245
    # pylint: disable=missing-function-docstring
    def write_line_break(self, data: Optional[str] = None):
        super().write_line_break(data)

        if len(self.indents) == 1:
            super().write_line_break()



# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import unittest

from click.testing import CliRunner
from mock import patch

from slo_generator.cli import main
from slo_generator.utils import load_config

from .test_stubs import CTX, mock_sd

cwd = os.path.dirname(os.path.abspath(__file__))
root = os.path.dirname(os.path.dirname(cwd))


class TestCLI(unittest.TestCase):
    def setUp(self):
        for key, value in CTX.items():
            os.environ[key] = value
        slo_config = f"{root}/samples/cloud_monitoring/slo_gae_app_availability.yaml"  # noqa: E501
        config = f"{root}/samples/config.yaml"
        self.slo_config = slo_config
        self.slo_metadata_name = load_config(slo_config, ctx=CTX)["metadata"]["name"]
        self.config = config
        self.cli = CliRunner()

    @patch("google.api_core.grpc_helpers.create_channel", return_value=mock_sd(8))
    def test_cli_compute(self, mock):
        args = ["compute", "-f", self.slo_config, "-c", self.config]
        result = self.cli.invoke(main, args)
        self.assertEqual(result.exit_code, 0)

    @patch("google.api_core.grpc_helpers.create_channel", return_value=mock_sd(40))
    def test_cli_compute_folder(self, mock):
        args = ["compute", "-f", f"{root}/samples/cloud_monitoring", "-c", self.config]
        result = self.cli.invoke(main, args)
        self.assertEqual(result.exit_code, 0)

    def test_cli_compute_no_config(self):
        args = ["compute", "-f", f"{root}/samples", "-c", f"{root}/samples/config.yaml"]
        result = self.cli.invoke(main, args)
        self.assertEqual(result.exit_code, 1)

    def test_cli_api(self):
        # TODO: Write test
        pass

    def test_cli_migrate(self):
        # TODO: Write test
        pass


if __name__ == "__main__":
    unittest.main()



"""
`open_search.py`
Opensearch backend implementation.
"""

import copy
import logging

from opensearchpy import OpenSearch

from slo_generator.constants import NO_DATA

LOGGER = logging.getLogger(__name__)


# pylint: disable=duplicate-code
class OpenSearchBackend:
    """Backend for querying metrics from OpenSearch.

    Args:
        client(opensearch.OpenSearch): Existing OS client.
        os_config(dict): OS client configuration.
    """

    def __init__(self, client=None, **os_config):
        self.client = client
        if self.client is None:
            conf = copy.deepcopy(os_config)
            url = conf.pop("url", None)
            basic_auth = conf.pop("basic_auth", None)
            api_key = conf.pop("api_key", None)
            if url:
                conf["hosts"] = url
            if basic_auth:
                conf["basic_auth"] = (basic_auth["username"], basic_auth["password"])
            if api_key:
                conf["api_key"] = (api_key["id"], api_key["value"])

            self.client = OpenSearch(**conf)

    # pylint: disable=unused-argument
    def good_bad_ratio(self, timestamp, window, slo_config):
        """Query two timeseries, one containing 'good' events, one containing
        'bad' events.

        Args:
            timestamp(int): UNIX timestamp.
            window(int): Window size (in seconds).
            slo_config(dict): SLO configuration.
              spec:
                method: "good_bad_ratio"
                service_level_indicator:
                  query_good(str): the search query to look for good events
                  query_bad(str): the search query to look for ba events
                  query_valid(str): the search query to look for valid events

        Returns:
            tuple: good_event_count, bad_event_count
        """
        measurement = slo_config["spec"]["service_level_indicator"]
        index = measurement["index"]
        query_good = measurement["query_good"]
        query_bad = measurement.get("query_bad")
        query_valid = measurement.get("query_valid")
        date_field = measurement.get("date_field")

        good = OS.build_query(query_good, window, date_field)
        bad = OS.build_query(query_bad, window, date_field)
        valid = OS.build_query(query_valid, window, date_field)

        good_events_count = OS.count(self.query(index, good))

        if query_bad is not None:
            bad_events_count = OS.count(self.query(index, bad))
        elif query_valid is not None:
            bad_events_count = OS.count(self.query(index, valid)) - good_events_count
        else:
            raise ValueError("`filter_bad` or `filter_valid` is required.")

        return good_events_count, bad_events_count

    def query(self, index, body):
        """Query Opensearch server.

        Args:
            index(str): Index to query.
            body(dict): Query body.

        Returns:
            dict: Response.
        """
        return self.client.search(index=index, body=body)

    @staticmethod
    def count(response):
        """Count event in opensearch response.

        Args:
            response(dict): Opensearch query response.

        Returns:
            int: Event count.
        """
        try:
            return response["hits"]["total"]["value"]
        except KeyError as exception:
            LOGGER.warning("Couldn't find any values in timeseries response")
            LOGGER.debug(exception, exc_info=True)
            return NO_DATA

    @staticmethod
    def build_query(query, window, date_field):
        """Build Opensearch query.

        Add window to existing query.
        Replace window for different error budget steps on-the-fly.

        Args:
            query(dict): Existing query body.
            window(int): Window in seconds.
            date_field(str): Field to filter time on

        Returns:
            dict: Query body with range clause added.
        """
        if query is None:
            return None
        body = {"query": {"bool": query}, "track_total_hits": True}
        range_query = {
            f"{date_field}": {
                "gte": f"now-{window}s/s",
                "lt": "now/s",
            }
        }

        if "filter" in body["query"]["bool"]:
            body["query"]["bool"]["filter"]["range"] = range_query
        else:
            body["query"]["bool"]["filter"] = {"range": range_query}

        return body


OS = OpenSearchBackend



# Copyright 2020 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from slo_generator.report import SLOReport

from .test_stubs import mock_slo_report


class TestReport(unittest.TestCase):
    def test_report_enough_events(self):
        report_cfg = mock_slo_report("enough_events")
        report = SLOReport(**report_cfg)
        self.assertTrue(report.valid)
        self.assertEqual(report.sli_measurement, 0.5)
        self.assertEqual(report.alert, True)

    def test_report_no_good_events(self):
        report_cfg = mock_slo_report("no_good_events")
        report = SLOReport(**report_cfg)
        self.assertTrue(report.valid)
        self.assertEqual(report.sli_measurement, 0)

    def test_report_no_bad_events(self):
        report_cfg = mock_slo_report("no_bad_events")
        report = SLOReport(**report_cfg)
        self.assertTrue(report.valid)
        self.assertEqual(report.sli_measurement, 1)

    def test_report_valid_sli_value(self):
        report_cfg = mock_slo_report("valid_sli_value")
        report = SLOReport(**report_cfg)
        self.assertTrue(report.valid)
        self.assertEqual(report.sli_measurement, report_cfg["backend"]["sli"])
        self.assertEqual(report.alert, False)

    def test_report_no_events(self):
        report_cfg = mock_slo_report("no_events")
        report = SLOReport(**report_cfg)
        self.assertFalse(report.valid)

    def test_report_no_good_bad_events(self):
        report_cfg = mock_slo_report("no_good_bad_events")
        report = SLOReport(**report_cfg)
        self.assertFalse(report.valid)

    def test_report_not_enough_events(self):
        report_cfg = mock_slo_report("not_enough_events")
        report = SLOReport(**report_cfg)
        self.assertFalse(report.valid)

    def test_report_no_sli_value(self):
        report_cfg = mock_slo_report("no_sli_value")
        report = SLOReport(**report_cfg)
        self.assertFalse(report.valid)

    def test_report_no_backend_response_sli(self):
        report_cfg = mock_slo_report("no_backend_response_sli")
        report = SLOReport(**report_cfg)
        self.assertFalse(report.valid)

    def test_report_no_backend_response_ratio(self):
        report_cfg = mock_slo_report("no_backend_response_ratio")
        report = SLOReport(**report_cfg)
        self.assertFalse(report.valid)

    def test_report_invalid_backend_response_type(self):
        report_cfg = mock_slo_report("invalid_backend_response_type")
        report = SLOReport(**report_cfg)
        self.assertFalse(report.valid)



# Copyright 2022 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# flake8: noqa

import unittest

from slo_generator.backends.splunk import SplunkBackend


class TestSplunkBackend(unittest.TestCase):
    def test_fix_search_prefix(search):
        search: str = "index=* status=200"
        fixed_search = "search " + search

        assert SplunkBackend.fix_search_prefix(search) == fixed_search
        assert SplunkBackend.fix_search_prefix(fixed_search) == fixed_search



# Copyright 2020 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`base.py`
Base exporter abstract classes.
"""
import logging
import warnings
from abc import ABCMeta, abstractmethod

LOGGER = logging.getLogger(__name__)

# Default metric labels exported by all metrics exporters
DEFAULT_METRIC_LABELS = [
    "error_budget_policy_step_name",
    "service_name",
    "feature_name",
    "slo_name",
    "metadata",
]

# Default metrics that are exported by metrics exporters.
DEFAULT_METRICS = [
    {
        "name": "error_budget_burn_rate",
        "description": "Speed at which the error budget is consumed.",
        "labels": DEFAULT_METRIC_LABELS,
    },
    {
        "name": "alerting_burn_rate_threshold",
        "description": "Error Budget burn rate threshold.",
        "labels": DEFAULT_METRIC_LABELS,
    },
    {
        "name": "events_count",
        "description": "Number of events",
        "labels": DEFAULT_METRIC_LABELS + ["good_events_count", "bad_events_count"],
    },
    {
        "name": "sli_measurement",
        "description": "Service Level Indicator.",
        "labels": DEFAULT_METRIC_LABELS,
    },
    {
        "name": "slo_target",
        "description": "Service Level Objective target.",
        "labels": DEFAULT_METRIC_LABELS,
    },
]


class MetricsExporter:  # pytype: disable=ignored-metaclass
    """Abstract class to export metrics to different backends. Common format
    for YAML configuration to configure which metrics should be exported."""

    __metaclass__ = ABCMeta  # pytype: disable=ignored-metaclass

    def export(self, data, **config):
        """Export metric data. Loops through metrics config and calls the child
        class `export_metric` method.

        Args:
            data (dict): SLO Report data.
            config (dict): Exporter config.

        Returns:
            list: List of exporter responses.
        """
        metrics = config.get("metrics", DEFAULT_METRICS)
        required_fields = getattr(self, "REQUIRED_FIELDS", [])
        optional_fields = getattr(self, "OPTIONAL_FIELDS", [])
        LOGGER.debug(f"Exporting {len(metrics)} metrics with {self.__class__.__name__}")
        for metric_cfg in metrics:
            if isinstance(metric_cfg, str):  # short form
                metric_cfg = {
                    "name": metric_cfg,
                    "alias": metric_cfg,
                    "description": "",
                    "labels": DEFAULT_METRIC_LABELS,
                }
            if metric_cfg["name"] == "error_budget_burn_rate":
                metric_cfg = MetricsExporter.use_deprecated_fields(
                    config=config, metric=metric_cfg
                )
            metric = metric_cfg.copy()
            fields = {
                key: value
                for key, value in config.items()
                if key in required_fields or key in optional_fields
            }
            metric.update(fields)
            metric = self.build_metric(data, metric)
            self.export_metric(metric)

    def build_metric(self, data, metric):
        """Build a metric from current data and metric configuration.
        Set the metric value labels and eventual alias.

        Args:
            data (dict): SLO Report data.
            metric (dict): Metric configuration.

        Returns:
            dict: Metric configuration.
        """
        name = metric["name"]
        prefix = getattr(self, "METRIC_PREFIX", None)

        # Set value + timestamp
        metric["value"] = data[name]
        metric["timestamp"] = data["timestamp"]

        # Set metric data labels
        labels = metric.get("labels", DEFAULT_METRIC_LABELS).copy()
        additional_labels = metric.get("additional_labels", [])
        labels.extend(additional_labels)
        labels = MetricsExporter.build_data_labels(data, labels)
        metric["labels"] = labels

        # Use metric alias (mapping)
        if "alias" in metric:
            metric["name"] = metric["alias"]

        if prefix and not metric["name"].startswith(prefix):
            metric["name"] = prefix + metric["name"]

        # Set description
        metric["description"] = metric.get("description", "")
        return metric

    @staticmethod
    def build_data_labels(data, labels):
        """Build data labels. Also handle nested labels (depth=1).

        Args:
            data (dict): SLO Report data.
            labels (list): Label keys.

        Returns:
            dict: Data labels.
        """
        data_labels = {}
        nested_labels = [
            label for label in labels if label in data and isinstance(data[label], dict)
        ]
        flat_labels = [
            label
            for label in labels
            if label in data and not isinstance(data[label], dict)
        ]
        for label in nested_labels:
            data_labels.update({k: str(v) for k, v in data[label].items()})
        for label in flat_labels:
            data_labels[label] = str(data[label])
        LOGGER.debug(f"Data labels: {data_labels}")
        return data_labels

    @staticmethod
    def use_deprecated_fields(config, metric):
        """Old format to new format with FutureWarning for 2.0.0.

        Update error_budget_burn_rate metric with `metric_type`,
        `metric_labels`, and `metric_description`.

        Args:
            config (dict): Exporter config.
            metric (dict): Metric config.

        Returns:
            list: List of metrics to export.
        """
        old_metric_type = config.get("metric_type")
        old_metric_labels = config.get("metric_labels")
        old_metric_description = config.get("metric_description")
        if old_metric_type:
            metric["alias"] = old_metric_type
            warnings.warn(
                "`metric_type` will be deprecated in favor of `metrics` "
                "in version 2.0.0, ",
                FutureWarning,
            )
        if old_metric_labels:
            metric["labels"] = old_metric_labels
            warnings.warn(
                "`metric_labels` will be deprecated in favor of `metrics` "
                "in version 2.0.0, ",
                FutureWarning,
            )
        if old_metric_description:
            warnings.warn(
                "`metric_description` will be deprecated in favor of `metrics` "
                "in version 2.0.0, ",
                FutureWarning,
            )
            metric["description"] = old_metric_description
        return metric

    @abstractmethod
    def export_metric(self, data):
        """Abstract method to export a metric. Should be implemented by children
        classes."""
        raise NotImplementedError



# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from slo_generator.utils import (
    get_backend_cls,
    get_exporter_cls,
    get_human_time,
    import_dynamic,
)


class TestUtils(unittest.TestCase):
    def test_get_human_time(self):
        # Timezones
        tz_UTC = "UTC"
        tz_Paris = "Europe/Paris"
        tz_Chicago = "America/Chicago"

        # Timestamp 1
        timestamp_20230615T16 = 1686845838
        utc_time_20230615T16 = "2023-06-15T16:17:18+00:00"
        paris_time_20230615T16 = "2023-06-15T18:17:18+02:00"
        chicago_time_20230615T16 = "2023-06-15T11:17:18-05:00"
        human_utc_20230615T16 = get_human_time(timestamp_20230615T16, timezone=tz_UTC)
        human_paris_20230615T16 = get_human_time(
            timestamp_20230615T16, timezone=tz_Paris
        )
        human_chicago_20230615T16 = get_human_time(
            timestamp_20230615T16, timezone=tz_Chicago
        )

        # Timestamp 2
        timestamp_20231215T16 = 1702660513.987654
        utc_time_20231215T16 = "2023-12-15T17:15:13.987654+00:00"
        paris_time_20231215T16 = "2023-12-15T18:15:13.987654+01:00"
        chicago_time_20231215T16 = "2023-12-15T11:15:13.987654-06:00"
        human_utc_20231215T16 = get_human_time(timestamp_20231215T16, timezone=tz_UTC)
        human_paris_20231215T16 = get_human_time(
            timestamp_20231215T16, timezone=tz_Paris
        )
        human_chicago_20231215T16 = get_human_time(
            timestamp_20231215T16, timezone=tz_Chicago
        )

        self.assertEqual(utc_time_20230615T16, human_utc_20230615T16)
        self.assertEqual(paris_time_20230615T16, human_paris_20230615T16)
        self.assertEqual(chicago_time_20230615T16, human_chicago_20230615T16)

        self.assertEqual(utc_time_20231215T16, human_utc_20231215T16)
        self.assertEqual(paris_time_20231215T16, human_paris_20231215T16)
        self.assertEqual(chicago_time_20231215T16, human_chicago_20231215T16)

    def test_get_backend_cls(self):
        res1 = get_backend_cls("CloudMonitoring")
        res2 = get_backend_cls("Prometheus")
        self.assertEqual(res1.__name__, "CloudMonitoringBackend")
        self.assertEqual(res1.__module__, "slo_generator.backends.cloud_monitoring")
        self.assertEqual(res2.__name__, "PrometheusBackend")
        self.assertEqual(res2.__module__, "slo_generator.backends.prometheus")
        with self.assertWarns(ImportWarning):
            get_backend_cls("UndefinedBackend")

    def test_get_backend_dynamic_cls(self):
        res1 = get_backend_cls("pathlib.Path")
        self.assertEqual(res1.__name__, "Path")
        self.assertEqual(res1.__module__, "pathlib")
        with self.assertWarns(ImportWarning):
            get_exporter_cls("foo.bar.DoesNotExist")

    def test_get_exporter_cls(self):
        res1 = get_exporter_cls("CloudMonitoring")
        res2 = get_exporter_cls("Pubsub")
        res3 = get_exporter_cls("Bigquery")
        self.assertEqual(res1.__name__, "CloudMonitoringExporter")
        self.assertEqual(res1.__module__, "slo_generator.exporters.cloud_monitoring")
        self.assertEqual(res2.__name__, "PubsubExporter")
        self.assertEqual(res2.__module__, "slo_generator.exporters.pubsub")
        self.assertEqual(res3.__name__, "BigqueryExporter")
        self.assertEqual(res3.__module__, "slo_generator.exporters.bigquery")
        with self.assertWarns(ImportWarning):
            get_exporter_cls("UndefinedExporter")

    def test_get_exporter_dynamic_cls(self):
        res1 = get_exporter_cls("pathlib.Path")
        self.assertEqual(res1.__name__, "Path")
        self.assertEqual(res1.__module__, "pathlib")
        with self.assertWarns(ImportWarning):
            get_exporter_cls("foo.bar.DoesNotExist")

    def test_import_dynamic(self):
        res1 = import_dynamic(
            "slo_generator.backends.cloud_monitoring",
            "CloudMonitoringBackend",
            prefix="backend",
        )
        res2 = import_dynamic(
            "slo_generator.exporters.cloud_monitoring",
            "CloudMonitoringExporter",
            prefix="exporter",
        )
        self.assertEqual(res1.__name__, "CloudMonitoringBackend")
        self.assertEqual(res2.__name__, "CloudMonitoringExporter")
        with self.assertWarns(ImportWarning):
            import_dynamic(
                "slo_generator.backends.unknown",
                "CloudMonitoringUnknown",
                prefix="unknown",
            )


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`cloudevents.py`
CloudEvents exporter class.
"""
import logging

import google.auth.transport.requests
import requests
from cloudevents.http import CloudEvent, to_structured
from google.oauth2.id_token import fetch_id_token

LOGGER = logging.getLogger(__name__)


# pylint: disable=too-few-public-methods
class CloudeventExporter:
    """Cloudevent exporter class.

    Args:
        client (obj, optional): Existing Datadog client to pass.
        service_url (str): Cloudevent receiver service URL.
    """

    REQUIRED_FIELDS = ["service_url"]
    OPTIONAL_FIELDS = ["auth"]

    def export(self, data, **config):
        """Export data as CloudEvent to an HTTP service receiving cloud events.

        Args:
            data (dict): Metric data.
            config (dict): Exporter config.
        """
        attributes = {
            "source": "https://github.com/cloudevents/spec/pull",
            "type": "com.google.slo_generator.slo_report",
        }
        event = CloudEvent(attributes, data)
        headers, data = to_structured(event)
        service_url = config["service_url"]
        if "auth" in config:
            auth = config["auth"]
            id_token = None
            if "token" in auth:
                id_token = auth["token"]
            elif auth.get("google_service_account_auth", False):  # Google oauth
                auth = google.auth.transport.requests.Request()
                id_token = fetch_id_token(auth, service_url)
            if id_token:
                headers["Authorization"] = f"Bearer {id_token}"
        resp = requests.post(
            service_url,
            headers=headers,
            data=data,
            timeout=10,
        )
        resp.raise_for_status()
        return resp



# Copyright 2020 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`datadog.py`
Datadog backend implementation.
"""

import logging
import pprint

import datadog

LOGGER = logging.getLogger(__name__)
logging.getLogger("datadog.api").setLevel(logging.ERROR)


class DatadogBackend:
    """Backend for querying metrics from Datadog.

    Args:
        client (obj, optional): Existing Datadog client to pass.
        api_key (str): Datadog API key.
        app_key (str): Datadog APP key.
        kwargs (dict): Extra arguments to pass to initialize function.
    """

    def __init__(self, client=None, api_key=None, app_key=None, **kwargs):
        self.client = client
        if not self.client:
            options = {"api_key": api_key, "app_key": app_key}
            options.update(kwargs)
            datadog.initialize(**options)
            self.client = datadog.api

    # pylint: disable=too-many-locals
    def good_bad_ratio(self, timestamp, window, slo_config):
        """Query SLI value from good and valid queries.

        Args:
            timestamp (int): UNIX timestamp.
            window (int): Window (in seconds).
            slo_config (dict): SLO configuration.

        Returns:
            tuple: Good event count, Bad event count.
        """
        measurement = slo_config["spec"]["service_level_indicator"]
        operator = measurement.get("operator", "sum")
        operator_suffix = measurement.get("operator_suffix", "as_count()")
        start = timestamp - window
        end = timestamp
        query_good = measurement["query_good"]

        if measurement.get("query_bad"):
            query = measurement.get("query_bad")
        elif measurement.get("query_valid"):
            query = measurement.get("query_valid")
        else:
            raise ValueError("One of `query_bad` or `query_valid` is required.")

        query_good = self._fmt_query(
            query_good,
            window,
            operator,
            operator_suffix,
        )

        good_event_query = self.client.Metric.query(
            start=start,
            end=end,
            query=query_good,
        )

        query = self._fmt_query(
            query,
            window,
            operator,
            operator_suffix,
        )

        event_query = self.client.Metric.query(
            start=start,
            end=end,
            query=query,
        )

        good_event_count = DatadogBackend.count(good_event_query)
        event_count = DatadogBackend.count(event_query)
        if measurement.get("query_valid"):
            event_count = event_count - good_event_count

        LOGGER.debug(f"Good events: {good_event_count} | " f"Bad events: {event_count}")

        return good_event_count, event_count

    def query_sli(self, timestamp, window, slo_config):
        """Query SLI value directly.

        Args:
            timestamp (int): UNIX timestamp.
            window (int): Window (in seconds).
            slo_config (dict): SLO configuration.

        Returns:
            float: SLI value.
        """
        measurement = slo_config["spec"]["service_level_indicator"]
        start = timestamp - window
        end = timestamp
        query = measurement["query"]
        query = self._fmt_query(query, window)
        response = self.client.Metric.query(start=start, end=end, query=query)
        LOGGER.debug(f"Result valid: {pprint.pformat(response)}")
        return DatadogBackend.count(response, average=True)

    def query_slo(self, timestamp, window, slo_config):
        """Query SLO value from a given Datadog SLO.

        Args:
            timestamp (int): UNIX timestamp.
            window (int): Window (in seconds).
            slo_config (dict): SLO configuration.

        Returns:
            tuple: Good event count, bad event count.
        """
        slo_id = slo_config["spec"]["service_level_indicator"]["slo_id"]
        from_ts = timestamp - window
        slo_data = self.client.ServiceLevelObjective.get(id=slo_id)
        LOGGER.debug(f"SLO data: {slo_id} | Result: {pprint.pformat(slo_data)}")
        data = self.client.ServiceLevelObjective.history(
            id=slo_id,
            from_ts=from_ts,
            to_ts=timestamp,
        )
        try:
            LOGGER.debug(f"Timeseries data: {slo_id} | Result: {pprint.pformat(data)}")
            good_event_count = data["data"]["series"]["numerator"]["sum"]
            valid_event_count = data["data"]["series"]["denominator"]["sum"]
            bad_event_count = valid_event_count - good_event_count
            return (good_event_count, bad_event_count)
        except KeyError as exception:  # monitor-based SLI
            sli_value = data["data"]["overall"]["sli_value"] / 100
            LOGGER.debug(exception)
            return sli_value

    @staticmethod
    def _fmt_query(query, window, operator=None, operator_suffix=None):
        """Format Datadog query:

        * If the Datadog expression has a `[window]` placeholder, replace it by
        the current window. Otherwise, append it to the expression.

        * If prefix / suffix operators are defined, apply them to the metric.

        * If labels are defined, append them to existing labels.

        Args:
            query (str): Original query in YAML config.
            window (int): Query window (in seconds).
            operator (str): Operator (e.g: sum, avg, median, ...)
            operator_suffix (str): Operator suffix (e.g: as_count(), ...)

        Returns:
            str: Formatted query.
        """
        query = query.strip()
        if operator:
            query = f"{operator}:{query}"
        if "[window]" in query:
            query = query.replace("[window]", f"{window}")
        if operator_suffix:
            query = f"{query}.{operator_suffix}"
        LOGGER.debug(f"Query: {query}")
        return query

    @staticmethod
    def count(response, average=False):
        """Count events in time series.

        Args:
            response (dict):  Datadog Metrics API response.
            average (bool): Take average of result.

        Returns:
            int: Event count.
        """
        try:
            values = []
            pointlist = response["series"][0]["pointlist"]
            for point in pointlist:
                value = point[1]
                if value is None:
                    continue
                values.append(value)
            if not values:
                raise IndexError
            if average:
                return sum(values) / len(values)
            return sum(values)
        except (IndexError, AttributeError) as exception:
            LOGGER.debug(exception)
            return 0  # no events in timeseries



"""
`splunk.py`
Query a splunk search to compute a SLI as a custom slo-generator backend
"""
import copy
import json
import logging
import re

import splunklib.client as splunk_client

LOGGER = logging.getLogger(__name__)


class SplunkBackend:
    """
    Queries data from a Splunk instance (On Premises or Cloud)
    and make SLO out of them
    """

    def __init__(self, client=None, **splunk_config):
        self.client = client
        conf = copy.deepcopy(splunk_config)
        host = conf.pop("host", None)
        port = int(conf.pop("port", 8089))
        token = conf.pop("token", None)
        user = conf.pop("user", None)
        password = conf.pop("password", None)
        if not self.client:
            if token is not None:
                # Create a Service instance and log in using a token
                self.client = splunk_client.connect(
                    host=host,
                    port=port,
                    splunkToken=token,
                )
            else:
                # Create a Service instance and log in using user/pwd
                self.client = splunk_client.connect(
                    host=host,
                    port=port,
                    username=user,
                    password=password,
                )

    def good_bad_ratio(self, timestamp, window, slo_config):
        """
        Query SLI value from good and valid queries.
        If both search_query_bad & search_query_valid are supplied,
           "bad" takes precedence over valid.
        Args:
            timestamp (int): UNIX timestamp.
            window (int): Window (in seconds).
            slo_config (dict): SLO configuration.
              spec:
                method: "good_bad_ratio"
                service_level_indicator:
                  search_query_good (str): the search query to loook for good events,
                                           must return a single row/column named "good"
                  search_query_bad (str): the search query to loook for bad events,
                                           must return a single row/column named "bad"
                  search_query_valid (str): the search query to loook for valid events,
                                           must return a single row/column named "valid"
        Returns:
            tuple: Good event count, Bad event count.
        """
        kwargs_search = {
            "earliest_time": f"-{window}s",
            "latest_time": timestamp,
            "output_mode": "json",
        }
        result_good = int(
            self.splunk_query(
                slo_config["spec"]["service_level_indicator"]["search_query_good"],
                "good",
                **kwargs_search,
            )
        )
        if "search_query_bad" in slo_config["spec"]["service_level_indicator"]:
            result_bad = int(
                self.splunk_query(
                    slo_config["spec"]["service_level_indicator"]["search_query_bad"],
                    "bad",
                    **kwargs_search,
                )
            )
        elif "search_query_valid" in slo_config["spec"]["service_level_indicator"]:
            result_bad = (
                int(
                    self.splunk_query(
                        slo_config["spec"]["service_level_indicator"][
                            "search_query_valid"
                        ],
                        "valid",
                        **kwargs_search,
                    )
                )
                - result_good
            )

        return (result_good, result_bad)

    def query_sli(self, timestamp, window, slo_config):
        """Query SLI value directly.
        Args:
            timestamp (int): UNIX timestamp.
            window (int): Window (in seconds).
            slo_config (dict): SLO configuration.
              spec:
                method: "query_sli"
                service_level_indicator:
                  search_query (str): the search query to run.
                                      must return a single row
                                      with at least a column named "sli"
        Returns:
            float: SLI value.
        """
        kwargs_search = {
            "earliest_time": f"-{window}s",
            "latest_time": timestamp,
            "output_mode": "json",
        }
        result = self.splunk_query(
            slo_config["spec"]["service_level_indicator"]["search_query"],
            "sli",
            **kwargs_search,
        )
        return result["sli"]

    @staticmethod
    def fix_search_prefix(search=""):
        """
        Splunk API search queries must start with "search"
        but people are used to the search bar of the UI which doesn't

        Args:
            search(string): the search to execute
        Returns:
            The same string prefixed with "search " if needed
        """
        if not re.search("^search ", search):
            search = f"search {search}"
        return search

    def splunk_query(self, search="", result_column="", **kwargs_search):
        """
        Cleanup and sent the search query to splunk
        and return the content of the first row of the choosen column

        Args:
            search(string): the search string to run against Splunk
            result_column(string): the column to look for in the results of the search
            kwargs_search(item): search parameters
                                 as described in the Splunk oneshot search API
        Returns
            The value of the first row of the results for the selected column
        """
        search_query = self.fix_search_prefix(search)
        result_json = self.client.jobs.oneshot(search_query, **kwargs_search)
        return json.loads(str(result_json))["results"][0][result_column]



# Copyright 2022 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from slo_generator.backends.elasticsearch import ElasticsearchBackend


class TestElasticsearchBackend(unittest.TestCase):
    def test_build_query_with_empty_query(self):
        query: dict = None
        window: int = 3600
        date_field: str = "@timestamp"
        enriched_query: dict = None
        assert (
            ElasticsearchBackend.build_query(query, window, date_field)
            == enriched_query
        )

    def test_build_query_with_simple_query_and_no_filter(self):
        query: dict = {
            "must": {
                "term": {
                    "name": "JAgOZE8",
                },
            },
        }
        window: int = 3600
        date_field: str = "@timestamp"
        enriched_query: dict = {
            "query": {
                "bool": {
                    "must": {
                        "term": {
                            "name": "JAgOZE8",
                        },
                    },
                    "filter": {
                        "range": {
                            "@timestamp": {
                                "gte": "now-3600s/s",
                                "lt": "now/s",
                            },
                        },
                    },
                },
            },
            "track_total_hits": True,
        }
        assert (
            ElasticsearchBackend.build_query(query, window, date_field)
            == enriched_query
        )

    def test_build_query_with_simple_query_and_existing_filter(self):
        query: dict = {
            "must": {
                "term": {
                    "name": "JAgOZE8",
                },
            },
            "filter": {
                "term": {
                    "grade": 2,
                },
            },
        }
        window: int = 3600
        date_field: str = "@timestamp"
        enriched_query: dict = {
            "query": {
                "bool": {
                    "must": {
                        "term": {
                            "name": "JAgOZE8",
                        },
                    },
                    "filter": {
                        "term": {
                            "grade": 2,
                        },
                        "range": {
                            "@timestamp": {
                                "gte": "now-3600s/s",
                                "lt": "now/s",
                            },
                        },
                    },
                },
            },
            "track_total_hits": True,
        }
        assert (
            ElasticsearchBackend.build_query(query, window, date_field)
            == enriched_query
        )

    def test_build_query_with_simple_query_and_existing_filter_with_range(self):
        query: dict = {
            "must": {
                "term": {
                    "name": "JAgOZE8",
                },
            },
            "filter": {
                "range": {
                    "@timestamp": {
                        "gte": "2015-01-01",  # should be replaced, so can be anything
                        "lt": "2015-01-03",  # should be replaced, so can be anything
                    },
                },
            },
        }
        window: int = 3600
        date_field: str = "@timestamp"
        enriched_query: dict = {
            "query": {
                "bool": {
                    "must": {
                        "term": {
                            "name": "JAgOZE8",
                        },
                    },
                    "filter": {
                        "range": {
                            "@timestamp": {
                                "gte": "now-3600s/s",
                                "lt": "now/s",
                            },
                        },
                    },
                },
            },
            "track_total_hits": True,
        }
        assert (
            ElasticsearchBackend.build_query(query, window, date_field)
            == enriched_query
        )



import unittest

from slo_generator.backends.open_search import OpenSearchBackend


class TestOpenSearchBackend(unittest.TestCase):
    assert 1 == 1

    def test_build_query_with_empty_query(self):
        assert OpenSearchBackend.build_query(None, 3600, "date") is None

    def test_build_query_with_simple_query_and_no_filter(self):
        query: dict = {
            "must": {
                "term": {
                    "status": "200",
                },
            },
        }

        enriched_query = {
            "query": {
                "bool": {
                    "must": {
                        "term": {
                            "status": "200",
                        },
                    },
                    "filter": {
                        "range": {
                            "date": {
                                "gte": "now-3600s/s",
                                "lt": "now/s",
                            },
                        },
                    },
                },
            },
            "track_total_hits": True,
        }

        assert OpenSearchBackend.build_query(query, 3600, "date") == enriched_query

    def test_build_query_with_simple_query_and_simple_filter(self):
        query: dict = {
            "must": {
                "term": {
                    "status": "200",
                },
            },
            "filter": {
                "term": {
                    "type": "get",
                },
            },
        }

        enriched_query = {
            "query": {
                "bool": {
                    "must": {
                        "term": {
                            "status": "200",
                        },
                    },
                    "filter": {
                        "term": {
                            "type": "get",
                        },
                        "range": {
                            "date": {
                                "gte": "now-3600s/s",
                                "lt": "now/s",
                            },
                        },
                    },
                },
            },
            "track_total_hits": True,
        }

        assert OpenSearchBackend.build_query(query, 3600, "date") == enriched_query

    def test_build_query_with_simple_query_and_existing_filter_with_range(self):
        query: dict = {
            "must": {
                "term": {
                    "status": "200",
                },
            },
            "filter": {
                "range": {
                    "date": {
                        "gte": "2023-08-28",
                        "lt": "2023-08-29",
                    },
                },
            },
        }

        enriched_query: dict = {
            "query": {
                "bool": {
                    "must": {
                        "term": {
                            "status": "200",
                        },
                    },
                    "filter": {
                        "range": {
                            "date": {
                                "gte": "now-3600s/s",
                                "lt": "now/s",
                            },
                        },
                    },
                },
            },
            "track_total_hits": True,
        }

        assert OpenSearchBackend.build_query(query, 3600, "date") == enriched_query



# Copyright 2022 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# flake8: noqa

import unittest

from slo_generator.backends.cloud_monitoring_mql import CloudMonitoringMqlBackend


class TestCloudMonitoringMqlBackend(unittest.TestCase):
    def test_enrich_query_with_time_horizon_and_period(self):
        timestamp: float = 1666995015.5144777  # = 2022/10/28 22:10:15.5144777
        window: int = 3600  # in seconds
        query: str = """fetch gae_app
| metric 'appengine.googleapis.com/http/server/response_count'
| filter resource.project_id == 'slo-generator-demo'
| filter
    metric.response_code == 429
    || metric.response_code == 200
"""

        enriched_query = """fetch gae_app
| metric 'appengine.googleapis.com/http/server/response_count'
| filter resource.project_id == 'slo-generator-demo'
| filter
    metric.response_code == 429
    || metric.response_code == 200
| group_by [] | within 3600s, d'2022/10/28 22:10:15' | every 3600s"""

        assert (
            CloudMonitoringMqlBackend.enrich_query_with_time_horizon_and_period(
                timestamp, window, query
            )
            == enriched_query
        )



# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`datadog.py`
Datadog exporter class.
"""
import logging

import datadog

from .base import MetricsExporter

LOGGER = logging.getLogger(__name__)
logging.getLogger("datadog.api").setLevel(logging.ERROR)

DEFAULT_API_HOST = "https://api.datadoghq.com"


# pylint: disable=too-few-public-methods
class DatadogExporter(MetricsExporter):
    """Datadog exporter class.

    Args:
        client (obj, optional): Existing Datadog client to pass.
        api_key (str): Datadog API key.
        app_key (str): Datadog APP key.
        kwargs (dict): Extra arguments to pass to initialize function.
    """

    REQUIRED_FIELDS = ["api_key", "app_key"]
    OPTIONAL_FIELDS = ["api_host"]

    def export_metric(self, data):
        """Export a metric to Datadog.

        Args:
            data (dict): Metric data.

        Raises:
            DatadogError (object): Datadog exception object.
        """
        options = {
            "api_key": data["api_key"],
            "app_key": data["app_key"],
            "api_host": data.get("api_host", DEFAULT_API_HOST),
        }
        datadog.initialize(**options)
        client = datadog.api
        timestamp = data["timestamp"]
        tags = data["labels"]
        name = data["name"]
        value = data["value"]
        return client.Metric.send(
            metric=name,
            points=[(timestamp, value)],
            tags=tags,
        )



# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`cloud_monitoring_mql.py`
Cloud Monitoring backend implementation with MQL (Monitoring Query Language).
"""
import logging
import pprint
import typing
import warnings
from collections import OrderedDict
from datetime import datetime, timezone
from typing import List, Optional, Tuple

from google.api.distribution_pb2 import Distribution
from google.cloud.monitoring_v3 import QueryTimeSeriesRequest
from google.cloud.monitoring_v3.services.query_service import QueryServiceClient
from google.cloud.monitoring_v3.services.query_service.pagers import (
    QueryTimeSeriesPager,
)
from google.cloud.monitoring_v3.types.metric import TimeSeries

from slo_generator.constants import NO_DATA

LOGGER = logging.getLogger(__name__)


class CloudMonitoringMqlBackend:
    """Backend for querying metrics from Cloud Monitoring with MQL.

    Args:
        project_id (str): Cloud Monitoring host project id.
        client (google.cloud.monitoring_v3.QueryServiceClient, optional):
            Existing Cloud Monitoring Query client. Initialize a new client
            if omitted.
    """

    def __init__(self, project_id: str, client=None):
        self.client = client
        if client is None:
            self.client = QueryServiceClient()
        self.parent = self.client.common_project_path(  # type: ignore[union-attr]
            project_id
        )

    def good_bad_ratio(
        self,
        timestamp: int,
        window: int,
        slo_config: dict,
    ) -> Tuple[int, int]:
        """Query two timeseries, one containing 'good' events, one containing
        'bad' events.

        Args:
            timestamp (int): UNIX timestamp.
            window (int): Window size (in seconds).
            slo_config (dict): SLO configuration.

        Returns:
            tuple: A tuple (good_event_count, bad_event_count)
        """
        measurement: dict = slo_config["spec"]["service_level_indicator"]
        filter_good: str = measurement["filter_good"]
        filter_bad: Optional[str] = measurement.get("filter_bad")
        filter_valid: Optional[str] = measurement.get("filter_valid")

        # Query 'good events' timeseries
        good_ts: List[TimeSeries] = self.query(timestamp, window, filter_good)
        good_event_count: int = CM.count(good_ts)

        # Query 'bad events' timeseries
        bad_event_count: int
        if filter_bad:
            bad_ts: List[TimeSeries] = self.query(timestamp, window, filter_bad)
            bad_event_count = CM.count(bad_ts)
        elif filter_valid:
            valid_ts: List[TimeSeries] = self.query(timestamp, window, filter_valid)
            bad_event_count = CM.count(valid_ts) - good_event_count
        else:
            raise ValueError("One of `filter_bad` or `filter_valid` is required.")

        LOGGER.debug(
            f"Good events: {good_event_count} | " f"Bad events: {bad_event_count}"
        )

        return good_event_count, bad_event_count

    # pylint: disable=too-many-locals,disable=unused-argument
    def distribution_cut(
        self,
        timestamp: int,
        window: int,
        slo_config: dict,
    ) -> Tuple[int, int]:
        """Query one timeseries of type 'exponential'.

        Args:
            timestamp (int): UNIX timestamp.
            window (int): Window size (in seconds).
            slo_config (dict): SLO configuration.

        Returns:
            tuple: A tuple (good_event_count, bad_event_count).
        """
        measurement: dict = slo_config["spec"]["service_level_indicator"]
        filter_valid: str = measurement["filter_valid"]
        threshold_bucket: int = int(measurement["threshold_bucket"])
        good_below_threshold: typing.Optional[bool] = measurement.get(
            "good_below_threshold", True
        )

        # Query 'valid' events
        series = self.query(timestamp, window, filter_valid)

        if not series:
            return NO_DATA, NO_DATA  # no timeseries

        distribution_value: Distribution = (
            series[0].point_data[0].values[0].distribution_value
        )
        bucket_counts: list = distribution_value.bucket_counts
        valid_events_count: int = distribution_value.count

        # Explicit the exponential distribution result
        count_sum: int = 0
        distribution = OrderedDict()
        for i, bucket_count in enumerate(bucket_counts):
            count_sum += bucket_count
            distribution[i] = {"count_sum": count_sum}
        LOGGER.debug(pprint.pformat(distribution))

        lower_events_count: int
        upper_events_count: int
        if len(distribution) - 1 < threshold_bucket:
            # maximum measured metric is below the cut after bucket number
            lower_events_count = valid_events_count
            upper_events_count = 0
        else:
            lower_events_count = distribution[threshold_bucket]["count_sum"]
            upper_events_count = valid_events_count - lower_events_count

        good_event_count: int
        bad_event_count: int
        if good_below_threshold:
            good_event_count = lower_events_count
            bad_event_count = upper_events_count
        else:
            good_event_count = upper_events_count
            bad_event_count = lower_events_count

        return good_event_count, bad_event_count

    def exponential_distribution_cut(self, *args, **kwargs) -> Tuple[int, int]:
        """Alias for `distribution_cut` method to allow for backwards
        compatibility.
        """
        warnings.warn(
            "exponential_distribution_cut will be deprecated in version 2.0, "
            "please use distribution_cut instead",
            FutureWarning,
        )
        return self.distribution_cut(*args, **kwargs)

    def query_sli(
        self,
        timestamp: int,  # pylint: disable=unused-argument
        window: int,
        slo_config: dict,
    ) -> float:
        """Query SLI value from a given MQL query.

        Args:
            timestamp (int): UNIX timestamp.
            window (int): Window (in seconds).
            slo_config (dict): SLO configuration.

        Returns:
            float: SLI value.
        """
        measurement: dict = slo_config["spec"]["service_level_indicator"]
        query: str = measurement["query"]
        series: List[TimeSeries] = self.query(timestamp, window, query)
        sli_value: float = series[0].point_data[0].values[0].double_value
        LOGGER.debug(f"SLI value: {sli_value}")
        return sli_value

    def query(self, timestamp: float, window: int, query: str) -> List[TimeSeries]:
        """Query timeseries from Cloud Monitoring using MQL.

        Args:
            timestamp (float): Current timestamp.
            window (int): Window size (in seconds).
            query (str): MQL query.
        Returns:
            list: List of timeseries objects.
        """
        # Enrich query to aggregate and reduce time series over target window.
        query_with_time_horizon_and_period: str = (
            self.enrich_query_with_time_horizon_and_period(timestamp, window, query)
        )
        request = QueryTimeSeriesRequest(
            {"name": self.parent, "query": query_with_time_horizon_and_period}
        )
        # fmt: off
        timeseries_pager: QueryTimeSeriesPager = (
            self.client.query_time_series(request)  # type: ignore[union-attr]
        )
        # fmt: on
        timeseries: List[TimeSeries] = list(timeseries_pager)
        LOGGER.debug(pprint.pformat(timeseries))
        return timeseries

    @staticmethod
    def enrich_query_with_time_horizon_and_period(
        timestamp: float,
        window: int,
        query: str,
    ) -> str:
        """Enrich MQL query with time period and horizon.
        Args:
            timestamp (float): UNIX timestamp.
            window (int): Query window (in seconds).
            query (str): Base query in YAML config.
        Returns:
            str: Enriched query.
        """
        # Python uses floating point numbers to represent time in seconds since the
        # epoch, in UTC, with decimal part representing nanoseconds.
        # MQL expects dates formatted like "%Y/%m/%d %H:%M:%S" or "%Y/%m/%d-%H:%M:%S".
        # Reference: https://cloud.google.com/monitoring/mql/reference#lexical-elements
        end_time_str: str = datetime.fromtimestamp(timestamp, tz=timezone.utc).strftime(
            "%Y/%m/%d %H:%M:%S"
        )
        query_with_time_horizon_and_period: str = (
            query
            + f"| group_by [] | within {window}s, d'{end_time_str}' | every {window}s"
        )
        return query_with_time_horizon_and_period

    @staticmethod
    def count(timeseries: List[TimeSeries]) -> int:
        """Count events in time series assuming it was aligned with ALIGN_SUM
        and reduced with REDUCE_SUM (default).

        Args:
            :obj:`monitoring_v3.TimeSeries`: Timeseries object.

        Returns:
            int: Event count.
        """
        try:
            return timeseries[0].point_data[0].values[0].int64_value
        except (IndexError, AttributeError) as exception:
            LOGGER.debug(exception, exc_info=True)
            return NO_DATA  # no events in timeseries


CM = CloudMonitoringMqlBackend



# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`custom_exporter.py`
Dummy sample of a custom exporter.
"""
import logging

from slo_generator.exporters.base import MetricsExporter

LOGGER = logging.getLogger(__name__)


class CustomMetricExporter(MetricsExporter):
    """Custom exporter for metrics."""

    def export_metric(self, data):
        """Export data to custom destination.

        Args:
            data (dict): Metric data.

        Returns:
            object: Custom destination response.
        """
        LOGGER.info(f"Metric data: {data}")
        return {
            "status": "ok",
            "code": 200,
        }


# pylint: disable=too-few-public-methods
class CustomSLOExporter:
    """Custom exporter for SLO data."""

    # pylint: disable=unused-argument
    def export(self, data, **config):
        """Export data to custom destination.

        Args:
            data (dict): SLO Report data.

        Returns:
            object: Custom destination response.
        """
        LOGGER.info(f"SLO data: {data}")
        return {
            "status": "ok",
            "code": 200,
        }



# Copyright 2020 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`constants.py`
Constants and environment variables used in `slo-generator`.
"""
import os
from typing import Dict, List, Tuple

# Compute
NO_DATA: int = -1
MIN_VALID_EVENTS: int = int(os.environ.get("MIN_VALID_EVENTS", "1"))

# Global
LATEST_MAJOR_VERSION: str = "v2"
COLORED_OUTPUT: int = int(os.environ.get("COLORED_OUTPUT", "0"))
DRY_RUN: bool = bool(int(os.environ.get("DRY_RUN", "0")))
DEBUG: int = int(os.environ.get("DEBUG", "0"))

# Exporters supporting v2 SLO report format
V2_EXPORTERS: Tuple[str, ...] = ("Pubsub", "Cloudevent")

# Config skeletons
CONFIG_SCHEMA: dict = {
    "backends": {},
    "exporters": {},
    "error_budget_policies": {},
}
SLO_CONFIG_SCHEMA: dict = {
    "apiVersion": "",
    "kind": "",
    "metadata": {},
    "spec": {
        "description": "",
        "backend": "",
        "method": "",
        "exporters": [],
        "service_level_indicator": {},
    },
}

# Providers that have changed with v2 YAML config format. This mapping helps
# migrate them to their updated names.
PROVIDERS_COMPAT: Dict[str, str] = {
    "Stackdriver": "CloudMonitoring",
    "StackdriverServiceMonitoring": "CloudServiceMonitoring",
}

# Fields that have changed name with v2 YAML config format. This mapping helps
# migrate them back to their former name, so that exporters are backward-
# compatible with v1.
METRIC_LABELS_COMPAT: Dict[str, str] = {
    "goal": "slo_target",
    "description": "slo_description",
    "error_budget_burn_rate_threshold": "alerting_burn_rate_threshold",
}

# Fields that used to be specified in top-level of YAML config are now specified
# in metadata fields. This mapping helps migrate them back to the top level when
# exporting reports, so that exporters are backward-compatible with v1.
METRIC_METADATA_LABELS_TOP_COMPAT: List[str] = [
    "service_name",
    "feature_name",
    "slo_name",
]


# Colors / Status
# pylint: disable=too-few-public-methods
class Colors:
    """Colors for console output."""

    HEADER: str = "\033[95m"
    OKBLUE: str = "\033[94m"
    OKGREEN: str = "\033[92m"
    WARNING: str = "\033[93m"
    FAIL: str = "\033[91m"
    ENDC: str = "\033[0m"
    BOLD: str = "\033[1m"
    UNDERLINE: str = "\033[4m"


GREEN: str = Colors.OKGREEN
RED: str = Colors.FAIL
ENDC: str = Colors.ENDC
BOLD: str = Colors.BOLD
WARNING: str = Colors.WARNING
FAIL: str = "❌"
SUCCESS: str = "✅"
RIGHT_ARROW: str = "➞"
